# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: config.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='config.proto',
  package='',
  serialized_pb=_b('\n\x0c\x63onfig.proto\"\x81\x02\n\nExperiment\x12\x0c\n\x04name\x18\x01 \x02(\t\x12)\n\x06\x64omain\x18\x02 \x03(\x0b\x32\x19.Experiment.ParameterSpec\x12.\n\x0b\x66idel_space\x18\x03 \x03(\x0b\x32\x19.Experiment.ParameterSpec\x1a\x89\x01\n\rParameterSpec\x12\x0c\n\x04name\x18\x01 \x02(\t\x12\x13\n\x04type\x18\x02 \x01(\t:\x05\x66loat\x12\x11\n\x03min\x18\x03 \x01(\x01:\x04-inf\x12\x10\n\x03max\x18\x04 \x01(\x01:\x03inf\x12\r\n\x03\x64im\x18\x05 \x01(\t:\x00\x12\x10\n\x06kernel\x18\x06 \x01(\t:\x00\x12\x0f\n\x05items\x18\x07 \x01(\t:\x00')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_EXPERIMENT_PARAMETERSPEC = _descriptor.Descriptor(
  name='ParameterSpec',
  full_name='Experiment.ParameterSpec',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='Experiment.ParameterSpec.name', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='type', full_name='Experiment.ParameterSpec.type', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=_b("float").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='min', full_name='Experiment.ParameterSpec.min', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=True, default_value=-1e10000,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='max', full_name='Experiment.ParameterSpec.max', index=3,
      number=4, type=1, cpp_type=5, label=1,
      has_default_value=True, default_value=1e10000,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='dim', full_name='Experiment.ParameterSpec.dim', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='kernel', full_name='Experiment.ParameterSpec.kernel', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='items', full_name='Experiment.ParameterSpec.items', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=137,
  serialized_end=274,
)

_EXPERIMENT = _descriptor.Descriptor(
  name='Experiment',
  full_name='Experiment',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='Experiment.name', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='domain', full_name='Experiment.domain', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='fidel_space', full_name='Experiment.fidel_space', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_EXPERIMENT_PARAMETERSPEC, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=17,
  serialized_end=274,
)

_EXPERIMENT_PARAMETERSPEC.containing_type = _EXPERIMENT
_EXPERIMENT.fields_by_name['domain'].message_type = _EXPERIMENT_PARAMETERSPEC
_EXPERIMENT.fields_by_name['fidel_space'].message_type = _EXPERIMENT_PARAMETERSPEC
DESCRIPTOR.message_types_by_name['Experiment'] = _EXPERIMENT

Experiment = _reflection.GeneratedProtocolMessageType('Experiment', (_message.Message,), dict(

  ParameterSpec = _reflection.GeneratedProtocolMessageType('ParameterSpec', (_message.Message,), dict(
    DESCRIPTOR = _EXPERIMENT_PARAMETERSPEC,
    __module__ = 'config_pb2'
    # @@protoc_insertion_point(class_scope:Experiment.ParameterSpec)
    ))
  ,
  DESCRIPTOR = _EXPERIMENT,
  __module__ = 'config_pb2'
  # @@protoc_insertion_point(class_scope:Experiment)
  ))
_sym_db.RegisterMessage(Experiment)
_sym_db.RegisterMessage(Experiment.ParameterSpec)


# @@protoc_insertion_point(module_scope)
